"""The base data models for completion in the commandline.

Contains:
    CompletionModel       -- A simple tree model based on Python data.
    CompletionItem        -- One item in the CompletionModel.
    CompletionFilterModel -- A QSortFilterProxyModel subclass for completions.
"""


from collections import OrderedDict

from PyQt5.QtCore import (QAbstractItemModel, Qt, QModelIndex, QVariant,
                          QSortFilterProxyModel, pyqtSignal)


class CompletionModel(QAbstractItemModel):

    """A simple tree model based on Python OrderdDict containing tuples.

    Used for showing completions later in the CompletionView.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._data = OrderedDict()
        self.parents = []
        self.root = CompletionItem([""] * 2)

    def removeRows(self, position=0, count=1, parent=QModelIndex()):
        """Remove rows from the model.

        Overrides QAbstractItemModel::removeRows.
        """
        node = self._node(parent)
        self.beginRemoveRows(parent, position, position + count - 1)
        node.children.pop(position)
        self.endRemoveRows()

    def _node(self, index):
        """Return the interal data representation for index.

        Returns the CompletionItem for index, or the root CompletionItem if the
        index was invalid.
        """
        if index.isValid():
            return index.internalPointer()
        else:
            return self.root

    def columnCount(self, parent=QModelIndex()):
        """Return the column count in the model.

        Overrides QAbstractItemModel::columnCount.
        """
        # pylint: disable=unused-argument
        return self.root.column_count()

    def data(self, index, role=Qt.DisplayRole):
        """Return the data for role/index as QVariant.

        Returns an invalid QVariant on error.
        Overrides QAbstractItemModel::data.
        """
        if not index.isValid():
            return QVariant()
        item = index.internalPointer()
        try:
            return QVariant(item.data(index.column(), role))
        except (IndexError, ValueError):
            return QVariant()

    def flags(self, index):
        """Return the item flags for index.

        Returns Qt.NoItemFlags on error.
        Overrides QAbstractItemModel::flags.
        """
        # FIXME categories are not selectable, but moving via arrow keys still
        # tries to select them
        if not index.isValid():
            return Qt.NoItemFlags
        flags = Qt.ItemIsEnabled
        if len(index.internalPointer().children) > 0:
            return flags
        else:
            return flags | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Return the header data for role/index as QVariant.

        Returns an invalid QVariant on error.
        Overrides QAbstractItemModel::headerData.
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return QVariant(self.root.data(section))
        return QVariant()

    def setData(self, index, value, role=Qt.EditRole):
        """Set the data for role/index to value.

        Returns True on success, False on failure.
        Overrides QAbstractItemModel::setData.
        """
        if not index.isValid():
            return False
        item = index.internalPointer()
        try:
            item.setdata(index.column(), value, role)
        except (IndexError, ValueError):
            return False
        self.dataChanged.emit(index, index)
        return True

    def index(self, row, column, parent=QModelIndex()):
        """Return the QModelIndex for row/column/parent.

        Returns an invalid QModelIndex on failure.
        Overrides QAbstractItemModel::index.
        """
        if (0 <= row < self.rowCount(parent) and
                0 <= column < self.columnCount(parent)):
            pass
        else:
            return QModelIndex()

        if not parent.isValid():
            parent_item = self.root
        else:
            parent_item = parent.internalPointer()

        child_item = parent_item.children[row]
        if child_item:
            return self.createIndex(row, column, child_item)
        else:
            return QModelIndex()

    def parent(self, index):
        """Return the QModelIndex of the parent of the object behind index.

        Returns an invalid QModelIndex on failure.
        Overrides QAbstractItemModel::parent.
        """
        if not index.isValid():
            return QModelIndex()
        item = index.internalPointer().parent
        if item == self.root or item is None:
            return QModelIndex()
        return self.createIndex(item.row(), 0, item)

    def rowCount(self, parent=QModelIndex()):
        """Return the rowCount (children count) for a parent.

        Uses the root frame if parent is invalid.
        Overrides QAbstractItemModel::data.
        """
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            pitem = self.root
        else:
            pitem = parent.internalPointer()

        return len(pitem.children)

    def sort(self, column, order=Qt.AscendingOrder):
        """Sort the data in column according to order.

        Raises NotImplementedError, should be overwritten in a superclass.
        Overrides QAbstractItemModel::sort.
        """
        raise NotImplementedError

    def init_data(self):
        """Initialize the Qt model based on the data in self._data."""
        for (cat, items) in self._data.items():
            newcat = CompletionItem([cat], self.root)
            self.root.children.append(newcat)
            for item in items:
                newitem = CompletionItem(item, newcat)
                newcat.children.append(newitem)

    def mark_all_items(self, needle):
        """Mark a string in all items (children of root-children).

        needle -- The string to mark.
        """
        for i in range(self.rowCount()):
            cat = self.index(i, 0)
            for k in range(self.rowCount(cat)):
                idx = self.index(k, 0, cat)
                old = self.data(idx).value()
                marks = self._get_marks(needle, old)
                self.setData(idx, marks, Qt.UserRole)

    def _get_marks(self, needle, haystack):
        """Return the marks for needle in haystack."""
        pos1 = pos2 = 0
        marks = []
        if not needle:
            return marks
        while True:
            pos1 = haystack.find(needle, pos2)
            if pos1 == -1:
                break
            pos2 = pos1 + len(needle)
            marks.append((pos1, pos2))
        return marks


class CompletionItem():
    """An item (row) in a CompletionModel."""

    parent = None
    children = None
    _data = None
    _marks = None

    def __init__(self, data, parent=None):
        """Constructor for CompletionItem.

        data   -- The data for the model, as tuple (columns).
        parent -- An optional parent item.
        """
        self.parent = parent
        self._data = data
        self.children = []
        self._marks = []

    def data(self, column, role=Qt.DisplayRole):
        """Get the data for role/column.

        Raise ValueError if the role is invalid.
        """
        if role == Qt.DisplayRole:
            return self._data[column]
        elif role == Qt.UserRole:
            return self._marks
        else:
            raise ValueError

    def setdata(self, column, value, role=Qt.DisplayRole):
        """Set the data for column/role to value.

        Raise ValueError if the role is invalid.
        """
        if role == Qt.DisplayRole:
            self._data[column] = value
        elif role == Qt.UserRole:
            self._marks = value
        else:
            raise ValueError

    def column_count(self):
        """Return the column count in the item."""
        return len(self._data)

    def row(self):
        """Return the row index (int) of the item, or 0 if it's a root item."""
        if self.parent:
            return self.parent.children.index(self)
        return 0


class CompletionFilterModel(QSortFilterProxyModel):

    """Subclass of QSortFilterProxyModel with custom sorting/filtering."""

    pattern_changed = pyqtSignal(str)
    _pattern = None

    def __init__(self, parent=None):
        super().__init__(parent)
        self.pattern = ''

    @property
    def pattern(self):
        """Getter for pattern."""
        return self._pattern

    @pattern.setter
    def pattern(self, val):
        """Setter for pattern.

        Invalidates the filter and emits pattern_changed.
        """
        self._pattern = val
        self.invalidate()
        self.pattern_changed.emit(val)

    def filterAcceptsRow(self, row, parent):
        """Custom filter implementation.

        Overrides QSortFilterProxyModel::filterAcceptsRow.

        row    -- The row of the item.
        parent -- The parent item QModelIndex.

        Returns True if self.pattern is contained in item, or if it's a root
        item (category). Else returns False.
        """
        if parent == QModelIndex():
            return True
        idx = self.sourceModel().index(row, 0, parent)
        data = self.sourceModel().data(idx).value()
        # TODO more sophisticated filtering
        if not self.pattern:
            return True
        return self.pattern in data

    def lessThan(self, lindex, rindex):
        """Custom sorting implementation.

        lindex -- The QModelIndex of the left item (*left* < right)
        rindex -- The QModelIndex of the right item (left < *right*)

        Prefers all items which start with self.pattern. Other than that, uses
        normal Python string sorting.
        """
        left = self.sourceModel().data(lindex).value()
        right = self.sourceModel().data(rindex).value()

        leftstart = left.startswith(self.pattern)
        rightstart = right.startswith(self.pattern)

        if leftstart and rightstart:
            return left < right
        elif leftstart:
            return True
        elif rightstart:
            return False
        else:
            return left < right

    def first_item(self):
        """Returns the first item in the model."""
        cat = self.index(0, 0)
        return self.index(0, 0, cat)

    def last_item(self):
        """Returns the last item in the model."""
        cat = self.index(self.rowCount() - 1, 0)
        return self.index(self.rowCount(cat) - 1, 0, cat)
